#include <pthread.h>
#include <errno.h>
#include <string.h>

#define MUTEX_KLUDGE

int pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *attr)
{
  memset(mutex, 0, sizeof(*mutex));
  return 0;
}

int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
  return 0;
}

/*
 * try to lock a mutex
 * the mutex variable cnt is a count of how many threads want this mutex
 * if it is 0, the mutex is free
 * if it is >0, someone owns the mutex
 */
int
pthread_mutex_trylock(pthread_mutex_t *mutex)
{
  return (__sync_bool_compare_and_swap(&mutex->cnt, 0, 1)) ? 0 : EBUSY;
}

/*
 * lock a mutex
 * we increment the mutex want count; if it is now 1, we're done, otherwise
 * we have to wait for it
 */
int
pthread_mutex_lock(pthread_mutex_t *mutex)
{
#ifdef MUTEX_KLUDGE
  if (pthread_mutex_trylock(mutex) == 0)
#else
  if (__addlock(&mutex->cnt, 1) == 1)
#endif
    {
      return 0;
    }

  /* failed to get the mutex */
  /* sleep on the queue */
  do {
#ifdef MUTEX_KLUDGE
    usleep(100);
#else
    _pthread_sleep(&mutex->queue);
#endif
  } while (0 != pthread_mutex_trylock(mutex));
  return 0;
}

/*
 * release the mutex
 * we atomically decrement the count of threads that want this
 * mutex; if it goes from 1->0 then we are done, otherwise we
 * have to wait one of the threads that was sleeping on the queue
 */
int
pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  if (__addlock(&mutex->cnt, -1) == 0)
    {
      return 0;
    }
  /* someone else was sleeping on the queue
     (or is about to... watch out for the race condition
     where this function is called during another thread's
     attempt to lock the mutex)
  */
#ifndef MUTEX_KLUDGE
  while (!_pthread_wake(&mutex->queue))
    pthread_yield();
#endif

  return 0;
}
